(ns com.aurellem.gb.vbm
  (:import java.io.File)
  (:import org.apache.commons.io.FileUtils)
  (:use com.aurellem.gb.gb-driver))

;;;;;;;;;;;;; read vbm file 

(def ^:dynamic *moves-cache*
     (File. user-home "proj/vba-clojure/moves/"))

(defn buttons [mask]
  (loop [buttons []
         masks (seq (dissoc button-code :listen))]
    (if (empty? masks) buttons
        (let [[button value] (first masks)]
          (if (not= 0x0000 (bit-and value mask))
            (recur (conj buttons button) (rest masks))
            (recur buttons (rest masks)))))))

(defn vbm-bytes [#^File vbm]
  (let [bytes (FileUtils/readFileToByteArray vbm)
        ints (int-array (count bytes))]
    (areduce bytes idx _ nil
             (aset ints idx
                   (bit-and 0xFF (aget bytes idx))))
    ints))

(def vbm-header-length 255)

(defn repair-vbm
  "Two 0's must be inserted after every reset."
  [vbm-masks]
  (loop [fixed []
         pending vbm-masks]
    (if (empty? pending) fixed
        (let [mask (first pending)]
          (if (not= 0x0000 (bit-and mask (button-code :restart)))
            (recur (conj fixed mask 0x0000 0x0000) (next pending))
            (recur (conj fixed mask) (next pending)))))))

(defn vbm-masks [#^File vbm]
  (repair-vbm
   (map (fn [[a b]]
          (+ (bit-shift-left a 8) b))
        (partition
         2 (drop vbm-header-length (vbm-bytes vbm))))))

(defn vbm-buttons [#^File vbm]
  (map buttons (vbm-masks vbm)))

(defn convert-buttons
  "To write a vbm file, we must remove the last two buttons after any
   reset event."
  [buttons]
  (loop [fixed []
         pending buttons]
    (if (empty? pending) fixed
        (let [mask (first pending)]
          (if (contains? (set (first pending)) :reset)
            (recur (conj fixed mask) (drop 3 pending))
            (recur (conj fixed mask) (next pending)))))))

(defn moves-filename [name]
  (File. *moves-cache* (format "%s.vbm" name)))

(defn read-moves [name]
  (let [target (moves-filename name)]
    (if (.exists target)
      (vbm-buttons target))))
;;;;;;;;;;;;;; write moves to vbm file


(def vbm-header
  (byte-array
   (map
    byte
    [86 66 77 26 1 0 0 0 105 74 88 79 89 1 0 0 0 0 0 0 0 1 2 112 0 0 0
     0 0 0 0 0 1 0 0 0 80 79 75 69 77 79 78 32 89 69 76 76 1 -105 124 4
     3 0 0 0 0 0 0 0 0 1 0 0 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 82 111 98 101 114 116 32 32 77 99 73 110 116 121 114 101 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95 95
     95 95 95 95])))

(def vbm-trailer
  (byte-array
   (map byte [0])))

(defn buttons->vbm-bytes [buttons]
  (let [bytes-in-ints
        (map button-mask (convert-buttons buttons))
        high-bits (map #(bit-shift-right (bit-and 0xFF00 %) 8)
                       bytes-in-ints)
        low-bits (map #(bit-and 0xFF %) bytes-in-ints)
        convert-byte (fn [i] (byte (if (>= i 128) (- i 256) i)))
        contents
        (byte-array
         (concat
          vbm-header
          (map convert-byte (interleave high-bits low-bits))
          vbm-trailer))]
    contents))
        
(defn write-moves! [moves name]
  (let [target (moves-filename name)]
    (clojure.java.io/copy (buttons->vbm-bytes moves) target)
    target))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(use 'clojure.java.shell)

(def vba-linux (File. user-home "bin/vba-linux"))

(defn play-vbm [#^File vbm]
  (.delete yellow-save-file)
  (if (.exists vbm)
    (sh (.getCanonicalPath vba-linux)
        (str "--playmovie=" (.getCanonicalPath vbm))
        (.getCanonicalPath yellow-rom-image)))
  nil)

